@@ -99,18 +99,8 @@ def setup_environment():
     # We don't ever want a pager
     compat.setenv('GIT_PAGER', '')
 
-    # Setup *SSH_ASKPASS
-    git_askpass = core.getenv('GIT_ASKPASS')
-    ssh_askpass = core.getenv('SSH_ASKPASS')
-    if git_askpass:
-        askpass = git_askpass
-    elif ssh_askpass:
-        askpass = ssh_askpass
-    elif sys.platform == 'darwin':
-        askpass = resources.package_command('ssh-askpass-darwin')
-    else:
-        askpass = resources.package_command('ssh-askpass')
-
+    # Setup the openssh askpass credentials helper.
+    askpass = _get_askpass()
     compat.setenv('GIT_ASKPASS', askpass)
     compat.setenv('SSH_ASKPASS', askpass)
 
@@ -147,6 +137,35 @@ def setup_environment():
     compat.setenv('GIT_MERGE_AUTOEDIT', 'no')
 
 
+def _get_askpass():
+    """Get a default askpass program appropriate for the current environment"""
+    git_askpass = core.getenv('GIT_ASKPASS')
+    ssh_askpass = core.getenv('SSH_ASKPASS')
+    if git_askpass:
+        return git_askpass
+    if ssh_askpass:
+        return ssh_askpass
+    if sys.platform == 'darwin':
+        return resources.package_command('ssh-askpass-darwin')
+
+    kde_askpass = core.find_executable('ksshaskpass')
+    gnome_askpass = core.find_executable('gnome-ssh-askpass')
+    if gnome_askpass is None:
+        gnome_askpass = '/usr/lib/openssh/gnome-ssh-askpass'
+
+    desktop_session = os.environ.get('DESKTOP_SESSION', 'unknown')
+    if desktop_session == 'gnome':
+        order = (gnome_askpass, kde_askpass)
+    else:
+        order = (kde_askpass, gnome_askpass)
+
+    for askpass in order:
+        if askpass and os.path.exists(askpass):
+            return askpass
+
+    return resources.package_command('ssh-askpass')
+
+
 def get_icon_themes(context):
     """Return the default icon theme names"""
     result = []
@@ -194,7 +213,10 @@ class ColaApplication:
             theme_str = self.context.cfg.get('cola.theme', default='default')
         theme = themes.find_theme(theme_str)
         self.theme = theme
-        self._app.setStyleSheet(theme.build_style_sheet(self._app.palette()))
+
+        bold_fonts = self.context.cfg.get('cola.boldfonts', default=False)
+        theme_stylesheet = theme.build_style_sheet(self._app.palette(), bold_fonts)
+        self._app.setStyleSheet(theme_stylesheet)
 
         is_macos_theme = theme_str.startswith('macos-')
         if is_macos_theme:
@@ -284,7 +306,7 @@ class ColaQApplication(QtWidgets.QApplication):
         view.save_state(settings=session)
 
 
-def process_args(args):
+def process_args(args, setup_repo=False):
     """Process and verify command-line arguments"""
     if args.version:
         # Accept 'git cola --version' or 'git cola version'
@@ -294,6 +316,15 @@ def process_args(args):
     # Handle session management
     restore_session(args)
 
+    # Initialize args.repo. If a repository was specified as a
+    # positional argument then we will use that value.
+    # If unspecified, the current directory is used.
+    if setup_repo and not args.repo:
+        if args.args and git.is_git_repository(args.args[0]):
+            args.repo = args.args.pop(0)
+    if not args.repo:
+        args.repo = core.getcwd()
+
     # Bail out if --repo is not a directory
     repo = core.decode(args.repo)
     if repo.startswith('file:'):
@@ -323,14 +354,22 @@ def restore_session(args):
         args.repo = session.repo
 
 
-def application_init(args, update=False, app_name='Git Cola', setup_worktree=True):
+def application_init(
+    args,
+    update=False,
+    app_name='Git Cola',
+    setup_worktree=True,
+    setup_repo=False,
+):
     """Parses the command-line arguments and starts git-cola"""
     # Ensure that we're working in a valid git repository.
     # If not, try to find one.  When found, chdir there.
     setup_environment()
-    process_args(args)
+    process_args(args, setup_repo=setup_repo)
 
     context = new_context(args, app_name=app_name)
+    enforce_single_instance(context)
+
     timer = context.timer
     timer.start('init')
 
@@ -368,6 +407,57 @@ def create_context():
     return new_context(args)
 
 
+def enforce_single_instance(context):
+    """Ensure that only a single instance of the application is running"""
+    if not context.args.single_instance:
+        return
+    window_id = context.app_name.replace(' ', '-')
+    semaphore = QtCore.QSystemSemaphore(window_id, 1)
+    semaphore.acquire()  # Prevent other instances from interacting with shared memory.
+
+    # We want to prevent the same instance from launching from the same directory.
+    # This is not foolproof, as users can still change repositories while the app
+    # is running, but this at least provides a workflow where users can launch
+    # the app from within their repository and we will prevent multiple instances
+    # from being launched from within that same repository.
+    shared_mem_id = window_id + '-'
+    current_dir_hash = utils.sha256hex(os.getcwd())
+    # Shared memory IDs can be max 30 characters on macOS.
+    remaining_bytes = 30 - len(shared_mem_id)
+    shared_mem_id += current_dir_hash[:-remaining_bytes]
+
+    # Shared memory may is not freed when the application terminates abnormally on
+    # Linux/UNIX so we workaround that by detaching here.
+    if not utils.is_win32():
+        fix_shared_mem = QtCore.QSharedMemory(shared_mem_id)
+        if fix_shared_mem.attach():
+            fix_shared_mem.detach()
+
+    # If the shared memory segment can be attached then we are already running.
+    context.shared_memory = QtCore.QSharedMemory(shared_mem_id)
+    if context.shared_memory.attach():
+        is_running = True
+    else:
+        # Create a one-byte shared memory segment.
+        context.shared_memory.create(1)
+        is_running = False
+
+    semaphore.release()
+
+    if is_running:
+        format_vars = {
+            'app_name': context.app_name,
+            'dirname': os.getcwd(),
+        }
+        QtWidgets.QMessageBox.warning(
+            None,
+            N_('%(app_name)s is already running') % format_vars,
+            N_('%(app_name)s is already running in %(dirname)s (--single-instance).')
+            % format_vars,
+        )
+        sys.exit(1)
+
+
 def application_run(context, view, start=None, stop=None):
     """Run the application main loop"""
     initialize_view(context, view)
@@ -421,7 +511,6 @@ def add_common_arguments(parser):
         '-r',
         '--repo',
         metavar='<repo>',
-        default=core.getcwd(),
         help='open the specified git repository',
     )
 
@@ -455,6 +544,14 @@ def add_common_arguments(parser):
         '--theme', metavar='<name>', default=None, help='specify a GUI theme name'
     )
 
+    # Allow only a single instance
+    parser.add_argument(
+        '-S',
+        '--single-instance',
+        action='store_true',
+        help='only allow a single instance to be running',
+    )
+
 
 def new_application(context, args):
     """Create a new ColaApplication"""
@@ -618,6 +715,7 @@ class ApplicationContext:
     def __init__(self, args):
         self.args = args
         self.app = None  # ColaApplication
+        self.shared_memory = None  # QSharedMemory
         self.command_bus = None  # cmd.CommandBus
         self.git = None  # git.Git
         self.cfg = None  # gitcfg.GitConfig
@@ -637,12 +735,34 @@ class ApplicationContext:
         self.notifier.setParent(view)
         self.command_bus = cmd.CommandBus(parent=view)
         self.runtask = qtutils.RunTask(parent=view)
+        self.notifier.command.connect(self._command, Qt.QueuedConnection)
+        self.notifier.critical.connect(self._critical, Qt.QueuedConnection)
+        self.notifier.information.connect(self._information, Qt.QueuedConnection)
+        self.notifier.log.connect(self._log, Qt.QueuedConnection)
+        self.notifier.ready.emit()
+
+    def _command(self, title, cmd, status, out, err):
+        Interaction.command(title, cmd, status, out, err)
+
+    def _critical(self, title, kwargs):
+        Interaction.critical(title, **kwargs)
+
+    def _information(self, title, kwargs):
+        Interaction.information(title, **kwargs)
+
+    def _log(self, message):
+        Interaction.log(message)
 
 
 class Notifier(QtCore.QObject):
     """Message bus for generic one-off notifications"""
 
+    command = Signal(str, str, int, str, str)
+    critical = Signal(object, object)
+    information = Signal(object, object)
+    log = Signal(object)
     message = Signal(object)
+    ready = Signal()
 
     def __init__(self, context, parent=None):
         super().__init__(parent)
@@ -662,6 +782,14 @@ class Notifier(QtCore.QObject):
 
         self.message.connect(listener, type=Qt.QueuedConnection)
 
+    def emit_log(self, message):
+        """Emit a log message"""
+        self.log.emit(message)
+
+    def git_cmd(self, message):
+        """Emit a log message"""
+        self.emit_log(f'[git] {message}')
+
 
 def find_git():
     """Return the path of git.exe, or None if we can't find it."""
